"""gptsum CLI implementation."""

import argparse
import sys
import uuid
from typing import List, Optional

import gptsum
import gptsum.checksum
import gptsum.gpt


def build_parser() -> argparse.ArgumentParser:
    """Build the CLI argument parser."""
    docstring_firstline = gptsum.__doc__.strip().splitlines()[0]
    docstring_suffix = docstring_firstline.split(":", 1)[1]
    description = docstring_suffix.strip().capitalize()
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("--version", action="version", version=gptsum.__version__)

    subparsers = parser.add_subparsers(title="subcommands")

    embed_parser = subparsers.add_parser(
        "embed",
        help="calculate and embed the image checksum as its GUID",
        description="Calculate and embed the checksum of a disk image as its GUID.",
    )
    embed_parser.set_defaults(func=embed)
    embed_parser.add_argument(
        "image",
        type=argparse.FileType("rb+"),
        help="disk image file",
        metavar="FILE",
    )

    verify_parser = subparsers.add_parser(
        "verify",
        help="verify disk image contents against the expected checksum (GUID)",
        description=(
            "Calculate the expected checksum of a disk image, "
            "then verify the embedded disk GUID against it. "
            "If there's a mismatch, the tool will exit with exit code 1."
        ),
    )
    verify_parser.set_defaults(func=verify)
    verify_parser.add_argument(
        "image",
        type=argparse.FileType("rb"),
        help="disk image file",
        metavar="FILE",
    )

    get_guid_parser = subparsers.add_parser(
        "get-guid",
        help="read a disk image GUID",
        description="Read and print the GPT label GUID from a disk image.",
    )
    get_guid_parser.set_defaults(func=get_guid)
    get_guid_parser.add_argument(
        "image",
        type=argparse.FileType("rb"),
        help="disk image file",
        metavar="FILE",
    )

    set_guid_parser = subparsers.add_parser(
        "set-guid",
        help="set a disk image GUID",
        description="Write a new label GUID to a disk image.",
    )
    set_guid_parser.set_defaults(func=set_guid)
    set_guid_parser.add_argument(
        "image",
        type=argparse.FileType("rb+"),
        help="disk image file",
        metavar="FILE",
    )
    set_guid_parser.add_argument(
        "guid",
        type=uuid.UUID,
        help="new label GUID",
        metavar="GUID",
    )

    calculate_guid_parser = subparsers.add_parser(
        "calculate-expected-guid",
        help="calculate and print the expected checksum (GUID) of an image",
        description=(
            "Calculate and print the expected checksum (GUID) of a disk image, "
            "without writing it to the file."
        ),
    )
    calculate_guid_parser.set_defaults(func=calculate_guid)
    calculate_guid_parser.add_argument(
        "image",
        type=argparse.FileType("rb"),
        help="disk image file",
        metavar="FILE",
    )

    return parser


def get_guid(ns: argparse.Namespace) -> None:
    """Execute the 'get-guid' subcommand."""
    with gptsum.gpt.GPTImage(fd=ns.image.fileno()) as image:
        image.validate()
        primary = image.read_primary_gpt_header()
        print("{}".format(primary.disk_guid))


def set_guid(ns: argparse.Namespace) -> None:
    """Execute the 'set-guid' subcommand."""
    with gptsum.gpt.GPTImage(fd=ns.image.fileno()) as image:
        image.validate()
        image.update_guid(ns.guid)


def calculate_guid(ns: argparse.Namespace) -> None:
    """Execute the 'calculate-expected-guid' subcommand."""
    with gptsum.gpt.GPTImage(fd=ns.image.fileno()) as image:
        image.validate()
        digest = gptsum.checksum.calculate(image)
        checksum_guid = gptsum.checksum.digest_to_guid(digest)
        print("{}".format(checksum_guid))


def embed(ns: argparse.Namespace) -> None:
    """Execute the 'embed' subcommand."""
    with gptsum.gpt.GPTImage(fd=ns.image.fileno()) as image:
        image.validate()
        digest = gptsum.checksum.calculate(image)
        checksum_guid = gptsum.checksum.digest_to_guid(digest)
        if checksum_guid != image.read_primary_gpt_header().disk_guid:
            image.update_guid(checksum_guid)


def verify(ns: argparse.Namespace) -> None:
    """Execute the 'verify' subcommand."""
    with gptsum.gpt.GPTImage(fd=ns.image.fileno()) as image:
        image.validate()
        primary = image.read_primary_gpt_header()
        digest = gptsum.checksum.calculate(image)
        checksum_guid = gptsum.checksum.digest_to_guid(digest)

        if primary.disk_guid != checksum_guid:
            sys.stderr.write(
                "Disk GUID doesn't match expected checksum, "
                "got {}, expected {}\n".format(primary.disk_guid, checksum_guid)
            )
            sys.stderr.flush()
            sys.exit(1)


def main(args: Optional[List[str]] = None) -> None:
    """Run the CLI program."""
    parser = build_parser()
    ns = parser.parse_args(args)

    func = getattr(ns, "func", None)
    if func:
        func(ns)
    else:
        parser.print_usage()



"""
gptsum: a tool to make disk images using GPT partitions self-verifiable.

Like `isomd5sum <https://github.com/rhinstaller/isomd5sum>`_.
"""

from typing import cast

try:
    import importlib.metadata as importlib_metadata
except ImportError:  # pragma: no cover
    import importlib_metadata  # type: ignore


__version__: str = cast(str, importlib_metadata.version(__name__))  # type: ignore



"""GPT disk image checksum calculation."""

import hashlib
import os
import uuid
from typing import Callable, List, Union, cast

from gptsum import gpt

ZERO_GUID = uuid.UUID(bytes=b"\0" * 16)


def hash_file(
    fn: Callable[[Union[bytes, memoryview]], None], fd: int, size: int, offset: int
) -> int:
    """Repeatedly call a function on a slice of a file."""
    buffsize = 128 * 1024
    done = 0

    if hasattr(os, "preadv"):
        preadv = cast(
            Callable[[int, List[bytearray], int], int],
            getattr(os, "preadv"),  # noqa: B009
        )

        buff = bytearray(buffsize)
        bufflist = [buff]
        view = memoryview(buff)

        while size > 0:
            n = preadv(fd, bufflist, offset)

            n = min(n, size)

            if n < buffsize:
                fn(view[:n])
            else:
                fn(view)

            done += n
            size -= n
            offset += n
    else:  # Python <= 3.6
        while size > 0:
            data = os.pread(fd, buffsize, offset)
            datasize = len(data)

            n = min(datasize, size)

            if n < datasize:
                fn(data[:n])
            else:
                fn(data)

            done += n
            size -= n
            offset += n

    return done


def calculate(image: gpt.GPTImage) -> bytes:
    """Calculate the 16-byte checksum of the given image."""
    fd = image.fileno()

    hasher = hashlib.blake2b(digest_size=16)
    offset = 0

    mbr = gpt.pread_all(fd, gpt.MBR_SIZE, 0)
    hasher.update(mbr)
    offset += len(mbr)

    primary = image.read_primary_gpt_header()
    primary_with_zero_guid = primary.with_new_guid(ZERO_GUID)
    hasher.update(primary_with_zero_guid.pack(override_crc32=0))
    offset += len(primary.pack())

    last = os.fstat(fd).st_size - gpt.GPT_HEADER_SIZE

    offset += hash_file(hasher.update, fd, last - offset, offset)

    assert offset == os.fstat(fd).st_size - gpt.GPT_HEADER_SIZE  # noqa: S101

    backup = image.read_backup_gpt_header()
    backup_with_zero_guid = backup.with_new_guid(ZERO_GUID)
    hasher.update(backup_with_zero_guid.pack(override_crc32=0))
    offset += len(backup.pack())

    assert offset == os.fstat(fd).st_size  # noqa: S101

    return hasher.digest()


def digest_to_guid(digest: bytes) -> uuid.UUID:
    """Convert a 16-byte digest into a GUID."""
    return uuid.UUID(bytes=digest)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gptsum']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.6,<0.9'],
 ':python_version < "3.8"': ['importlib-metadata>=3.7.3,<4.0.0']}

entry_points = \
{'console_scripts': ['gptsum = gptsum.cli:main']}

setup_kwargs = {
    'name': 'gptsum',
    'version': '0.0.6',
    'description': 'A tool to make disk images using GPT partitions self-verifiable',
    'long_description': 'gptsum\n======\nA tool to make disk images using GPT_ partitions self-verifiable, like\n`isomd5sum`_.\n\nNote this *only* works for read-only, immutable images!\n\n.. _GPT: https://en.wikipedia.org/wiki/GUID_Partition_Table\n.. _isomd5sum: https://github.com/rhinstaller/isomd5sum\n\nQuickstart\n**********\n.. When making changes to the quickstart code below,\n   make sure \'tests/test_readme.py\' is updated accordingly.\n\nFirst, create an empty disk image::\n\n    $ truncate -s64M image.raw\n\nThen, create a GPT partition table in it. Note, below we set an explicit\n`label-id`. In general, this should not be done and a random GUID will be\ngenerated. Simply remove the line.\n\n::\n\n    $ sfdisk image.raw << EOF\n    label: gpt\n    label-id: 132e3631-1ec9-4411-ab25-9b95b54b0903\n    first-lba: 2048\n    EOF\n\n    Checking that no-one is using this disk right now ... OK\n\n    Disk image.raw: 64 MiB, 67108864 bytes, 131072 sectors\n    Units: sectors of 1 * 512 = 512 bytes\n    Sector size (logical/physical): 512 bytes / 512 bytes\n    I/O size (minimum/optimal): 512 bytes / 512 bytes\n\n    >>> Script header accepted.\n    >>> Script header accepted.\n    >>> Script header accepted.\n    >>> Done.\n    Created a new GPT disklabel (GUID: 132E3631-1EC9-4411-AB25-9B95B54B0903).\n\n    New situation:\n    Disklabel type: gpt\n    Disk identifier: 132E3631-1EC9-4411-AB25-9B95B54B0903\n\n    The partition table has been altered.\n    Syncing disks.\n\nWe can retrieve the current disk GUID using ``gptsum``::\n\n    $ gptsum get-guid image.raw\n    132e3631-1ec9-4411-ab25-9b95b54b0903\n\nVerification should fail::\n\n    $ gptsum verify image.raw || echo "Verification failed!"\n    Disk GUID doesn\'t match expected checksum, got 132e3631-1ec9-4411-ab25-9b95b54b0903, expected 6190f5bb-1967-14ec-9fbd-a7d213a45461\n    Verification failed!\n\nEmbed the disk checksum as the label GUID::\n\n    $ gptsum embed image.raw\n\nVerification should now succeed::\n\n    $ gptsum verify image.raw && echo "Verification succeeded!"\n    Verification succeeded!\n\nIndeed, the GUID was changed::\n\n    $ gptsum get-guid image.raw\n    6190f5bb-1967-14ec-9fbd-a7d213a45461\n\n\nHow It Works\n************\nGenerally, when checksums are used to validate the integrity of a file, this\nchecksum needs to be provided out-of-band, e.g., in a separate file. This\ncomplicates the process a bit, since now multiple files need to be kept around\n(and potentially in sync).\n\nBeing able to verify a file\'s integrity without any external information would\nbe great. However, if we embed the checksum of a file in the file itself, we\nchange its checksum and verification would fail. So we need to apply some\ntricks.\n\nThe `isomd5sum`_ tool is often used to verify the integrity of ISO files, e.g.,\nLinux distribution releases. It uses an unused location in the ISO9660 file\nformat to embed an MD5 checksum of the actual data segments of said file. As\nsuch, this MD5 checksum does not represent the complete file contents but only\nthe pieces of data we\'re interested in.\n\nWe can translate this to GPT-partitioned disk images as well. In the GPT\nformat, there\'s no room to embed any arbitrary blobs (unless we\'d use the\nreserved or padding sections of the headers, which should be zeroed out so we\nshouldn\'t). However, GPT disks are identified by a GUID which is stored in the\ntwo metadata sections stored on the disk, at LBA 1 and as the last LBA on the\ndisk (the so-called primary and secondary GPT headers). This leaves room for 16\nbytes of semi-arbitrary data.\n\nFurthermore, the GPT headers themselves, including the GUID, are protected\nusing a CRC32 checksum.\n\nAs such, we can apply the following procedure:\n\n- Zero out the CRC32 and GUID fields in both GPT headers\n- Calculate a 16 byte checksum of the resulting image (covering all data,\n  except for the CRC32 and GUID fields)\n- Embed the checksum as the GUID field in both GPT headers (now becoming the\n  disk GUID)\n- Update the CRC32 fields in both GPT headers\n\nAt this point we have a fully valid GPT disk image with a GUID representing\nthe actual data contained in the image. One could argue this is no longer a\nvalid GUID (indeed it\'s not), but since it\'s generated using a secure hashing\nalgorithm over a (potentially large) file, we can assume the entropy is\nsufficient to avoid collisions. Essentially, if two disk images were to get the\nsame GUID, they\'re very likely the exact same disk image, content-wise.\n\nVerifying an image file is roughly the same procedure:\n\n- Zero out the CRC32 and GUID fields in both GPT headers (in-memory)\n- Calculate a 16 byte checksum of the resulting image\n- Verify this calculated checksum equals the actual GUID embedded in the image\n\nImplementation Details\n**********************\n`gptsum` is implemented in Python, and compatible with Python 3.6 and later.\nIt uses the `Blake2b`_ checksum algorithm to construct a 16 byte digest\nof the disk image.\n\nVarious subcommands are exposed by the CLI, refer to the `documentation`_\nfor more details.\n\n.. _Blake2b: https://en.wikipedia.org/wiki/BLAKE_(hash_function)#BLAKE2\n.. _documentation: https://nicolast.github.io/gptsum/\n',
    'author': 'Nicolas Trangez',
    'author_email': 'ikke@nicolast.be',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NicolasT/gptsum',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Code to work with `GPT`_ images and headers.

.. _GPT: https://en.wikipedia.org/wiki/GUID_Partition_Table
"""

import binascii
import dataclasses
import os
import struct
import uuid
from pathlib import Path
from types import TracebackType
from typing import ContextManager, Optional, Type

LBA_SIZE = 512
MBR_SIZE = LBA_SIZE
GPT_HEADER_SIZE = LBA_SIZE

_EXPECTED_ACTUAL_HEADER_SIZE = 92
_EXPECTED_PADDING_SIZE = GPT_HEADER_SIZE - _EXPECTED_ACTUAL_HEADER_SIZE

_GPT_ACTUAL_HEADER_STRUCT_FORMAT = "<8s4sII4sQQQQ16sQIII"
_GPT_ACTUAL_HEADER_STRUCT = struct.Struct(_GPT_ACTUAL_HEADER_STRUCT_FORMAT)
assert _GPT_ACTUAL_HEADER_STRUCT.size == _EXPECTED_ACTUAL_HEADER_SIZE  # noqa: S101
_GPT_HEADER_STRUCT = struct.Struct(
    "{}{}s".format(_GPT_ACTUAL_HEADER_STRUCT_FORMAT, _EXPECTED_PADDING_SIZE)
)
assert _GPT_HEADER_STRUCT.size == GPT_HEADER_SIZE  # noqa: S101

_EXPECTED_SIGNATURE = b"EFI PART"
_EXPECTED_REVISION = b"\0\0\1\0"
_EXPECTED_RESERVED = b"\0" * 4
_EXPECTED_PADDING = b"\0" * _EXPECTED_PADDING_SIZE


class GPTError(Exception):
    """Generic GPT error."""


class InvalidHeaderSizeError(ValueError, GPTError):
    """Header of invalid size passed in."""


class InvalidFieldError(ValueError, GPTError):
    """A GPT header field has invalid contents."""


class InvalidSignatureError(InvalidFieldError):
    """Invalid signature found in header."""


class UnsupportedRevisionError(ValueError, GPTError):
    """GPT header of unsupported revision passed in."""


class HeaderChecksumMismatchError(ValueError, GPTError):
    """GPT header checksum mismatch."""


class InvalidImageError(ValueError, GPTError):
    """A file which is supposedly a GPT image is not."""


@dataclasses.dataclass(frozen=True)
class GPTHeader(object):
    """Representation of a GPT header."""

    current_lba: int
    backup_lba: int
    first_usable_lba: int
    last_usable_lba: int
    disk_guid: uuid.UUID
    entries_starting_lba: int
    num_entries: int
    entry_size: int
    entries_crc32: int

    @classmethod
    def unpack(cls, raw_header: bytes) -> "GPTHeader":
        """Unpack a GPT header from its raw encoding.

        The given raw header must include all padding, i.e., its length must be
        512 bytes.

        :param raw_header: Raw GPT header
        :type raw_header: bytes

        :return: A :class:`GPTHeader` instance representing the raw GPT header
        :rtype: GPTHeader

        :raises InvalidHeaderSizeError: Given `raw_header` has invalid length
        :raises InvalidSignatureError: Invalid GPT signature in raw header
        :raises UnsupportedRevisionError: Unsupported GPT revision
        :raises InvalidFieldError: Invalid GPT field value detected
        :raises HeaderChecksumMismatchError: Header checksum mismatch detected
        """
        if len(raw_header) != GPT_HEADER_SIZE:
            raise InvalidHeaderSizeError(
                "Not a valid GPT header, must be {} bytes".format(GPT_HEADER_SIZE)
            )

        (
            signature,
            revision,
            header_size,
            header_crc32,
            reserved,
            current_lba,
            backup_lba,
            first_usable_lba,
            last_usable_lba,
            disk_guid,
            entries_starting_lba,
            num_entries,
            entry_size,
            entries_crc32,
            padding,
        ) = _GPT_HEADER_STRUCT.unpack(raw_header)

        if signature != _EXPECTED_SIGNATURE:
            raise InvalidSignatureError(
                "Not a valid GPT header, signature must be {}".format(
                    _EXPECTED_SIGNATURE.decode("ascii")
                )
            )

        if revision != _EXPECTED_REVISION:
            raise UnsupportedRevisionError("Unsupported GPT revision")

        if header_size != _EXPECTED_ACTUAL_HEADER_SIZE:
            raise InvalidHeaderSizeError(
                "GPT header is supposedly {} bytes long, expected {}".format(
                    header_size, _EXPECTED_ACTUAL_HEADER_SIZE
                )
            )

        header = raw_header[:header_size]

        if reserved != _EXPECTED_RESERVED:
            raise InvalidFieldError("GPT 'reserved' field has unexpected contents")

        if padding != _EXPECTED_PADDING:
            raise InvalidFieldError("GPT header padding has unexpected contents")

        prefix_crc32 = binascii.crc32(header[: 8 + 4 + 4])
        with_zeroed_crc32_crc32 = binascii.crc32(b"\0" * 4, prefix_crc32)
        calculated_header_crc32 = binascii.crc32(
            header[8 + 4 + 4 + 4 :], with_zeroed_crc32_crc32
        )
        calculated_header_crc32 &= 0xFFFFFFFF

        if header_crc32 != calculated_header_crc32:
            raise HeaderChecksumMismatchError(
                "GPT header CRC32 mismatch, got {}, expected {}".format(
                    calculated_header_crc32, header_crc32
                )
            )

        return cls(
            current_lba,
            backup_lba,
            first_usable_lba,
            last_usable_lba,
            uuid.UUID(bytes_le=disk_guid),
            entries_starting_lba,
            num_entries,
            entry_size,
            entries_crc32,
        )

    def pack(self, override_crc32: Optional[int] = None) -> bytes:
        """Pack the header in its raw bytes serialized form.

        :param override_crc32: Use the given value as CRC32 instead of calculating it

        :return: Serialized representation of the GPT header
        :rtype: bytes
        """
        if override_crc32 is not None:
            header_crc32 = override_crc32
        else:
            header_with_zero_crc32 = _GPT_ACTUAL_HEADER_STRUCT.pack(
                _EXPECTED_SIGNATURE,
                _EXPECTED_REVISION,
                _EXPECTED_ACTUAL_HEADER_SIZE,
                0,
                _EXPECTED_RESERVED,
                self.current_lba,
                self.backup_lba,
                self.first_usable_lba,
                self.last_usable_lba,
                self.disk_guid.bytes_le,
                self.entries_starting_lba,
                self.num_entries,
                self.entry_size,
                self.entries_crc32,
            )
            header_crc32 = binascii.crc32(header_with_zero_crc32)

        return _GPT_HEADER_STRUCT.pack(
            _EXPECTED_SIGNATURE,
            _EXPECTED_REVISION,
            _EXPECTED_ACTUAL_HEADER_SIZE,
            header_crc32,
            _EXPECTED_RESERVED,
            self.current_lba,
            self.backup_lba,
            self.first_usable_lba,
            self.last_usable_lba,
            self.disk_guid.bytes_le,
            self.entries_starting_lba,
            self.num_entries,
            self.entry_size,
            self.entries_crc32,
            _EXPECTED_PADDING,
        )

    def is_backup_of(self, other: "GPTHeader") -> bool:
        """Check whether the given GPT header is a backup of this one."""
        return all(
            [
                self.current_lba == other.backup_lba,
                self.backup_lba == other.current_lba,
                self.first_usable_lba == other.first_usable_lba,
                self.last_usable_lba == other.last_usable_lba,
                self.disk_guid == other.disk_guid,
                self.num_entries == other.num_entries,
                self.entry_size == other.entry_size,
                self.entries_crc32 == other.entries_crc32,
            ]
        )

    def with_new_guid(self, new_guid: uuid.UUID) -> "GPTHeader":
        """Construct a copy of the GPT header with a new GUID."""
        return GPTHeader(
            self.current_lba,
            self.backup_lba,
            self.first_usable_lba,
            self.last_usable_lba,
            new_guid,
            self.entries_starting_lba,
            self.num_entries,
            self.entry_size,
            self.entries_crc32,
        )


def pread_all(fd: int, size: int, offset: int) -> bytes:
    """Use :func:`os.pread` to read data, handling partial reads."""
    pieces = []
    while size > 0:
        read = os.pread(fd, size, offset)
        pieces.append(read)
        len_read = len(read)
        size -= len_read
        offset += len_read
    return b"".join(pieces)


def pwrite_all(fd: int, data: bytes, offset: int) -> None:
    """Use :func:`os.pwrite` to write data, handling partial writes."""
    while len(data) != 0:
        written = os.pwrite(fd, data, offset)
        data = data[written:]
        offset += written


class GPTImage(ContextManager["GPTImage"]):
    """Wrapper around a GPT-partitioned disk image file."""

    def __init__(
        self,
        *,
        fd: Optional[int] = None,
        path: Optional[Path] = None,
        open_mode: int = os.O_RDWR,
    ):
        """Construct a new :class:`GPTImage`.

        :param fd: File descriptor to open image file
        :param path: Path to image file to use
        :param open_mode: Mode to use when opening the disk image file

        :raises ValueError: Both or none of `fd` and `path` are provided
        """
        if fd is None and path is None:
            raise ValueError("Either fd or path must be given")

        if fd is not None and path is not None:
            raise ValueError("Both fd and path can't be given")

        self._fd = fd
        self._path = path
        self._open_mode = open_mode

    def fileno(self) -> int:
        """Get a file descriptor for the image file.

        :returns: FD to the image file

        :raises RuntimeError: No open FD to the image
        """
        if self._fd is None:
            raise RuntimeError("No open file descriptor")  # pragma: no cover

        return self._fd

    def __enter__(self) -> "GPTImage":
        """Open the disk image for use in a context.

        If `fd` was passed during construction, nothing happens, and the FD won't be
        closed by `__exit__`.

        :returns: `self`

        :raises InvalidImageError: File is not a valid GPT image
        """
        if self._fd is None:
            assert self._path is not None  # noqa: S101
            self._fd = os.open(
                self._path, os.O_CLOEXEC | os.O_LARGEFILE | self._open_mode
            )

        if os.fstat(self._fd).st_size < MBR_SIZE + GPT_HEADER_SIZE + GPT_HEADER_SIZE:
            if self._path is not None:
                os.close(self._fd)
                self._fd = None
            raise InvalidImageError("Image file too small to be a valid GPT image")

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Close the disk image when exiting a context."""
        if self._path is not None and self._fd is not None:
            os.close(self._fd)
            self._fd = None

    def _read_gpt_header(self, offset: int) -> GPTHeader:
        if self._fd is None:
            raise RuntimeError("No open file descriptor")  # pragma: no cover
        raw_header = pread_all(self._fd, GPT_HEADER_SIZE, offset)
        return GPTHeader.unpack(raw_header)

    def read_primary_gpt_header(self) -> GPTHeader:
        """Read the primary GPT header from the image."""
        return self._read_gpt_header(MBR_SIZE)

    def read_backup_gpt_header(self) -> GPTHeader:
        """Read the backup GPT header from the image."""
        if self._fd is None:
            raise RuntimeError("No open file descriptor")  # pragma: no cover
        size = os.fstat(self._fd).st_size
        return self._read_gpt_header(size - GPT_HEADER_SIZE)

    def validate(self) -> None:
        """Validate the GPT headers found in the image.

        :raises InvalidImageError: GPT headers are not compatible
        """
        gpt1 = self.read_primary_gpt_header()
        gpt2 = self.read_backup_gpt_header()

        if not gpt2.is_backup_of(gpt1) or not gpt1.is_backup_of(gpt2):
            raise InvalidImageError("GPT headers don't match")

    def write_gpt_headers(self, primary: GPTHeader, backup: GPTHeader) -> None:
        """Write primary and backup GPT headers to the image.

        :param primary: Primary GPT header to write
        :param backup: Backup GPT header to write

        :raises RuntimeError: No open file descriptor
        :raises ValueError: Given headers are not backups of each other
        :raises ValueError: Primary header has invalid 'current_lba'
        :raises ValueError: Backup header has invalid 'current_lba'
        """
        if self._fd is None:
            raise RuntimeError("No open file descriptor")  # pragma: no cover

        if not backup.is_backup_of(primary) or not primary.is_backup_of(backup):
            raise ValueError("Given headers are not backups of each other")

        if primary.current_lba != 1:
            raise ValueError("Primary header has invalid 'current_lba', expected 1")

        size = os.fstat(self._fd).st_size
        expected_backup_lba = size // 512 - 1
        if backup.current_lba != expected_backup_lba:
            raise ValueError(
                "Backup header has invalid 'current_lba', expected {}".format(
                    expected_backup_lba
                )
            )

        gpt1 = primary.pack()
        gpt2 = backup.pack()

        pwrite_all(self._fd, gpt1, primary.current_lba * LBA_SIZE)
        pwrite_all(self._fd, gpt2, backup.current_lba * LBA_SIZE)
        os.fsync(self._fd)

    def update_guid(self, guid: uuid.UUID) -> None:
        """Update the label GUID of the image to some new UUID.

        The new GPT headers are written to the image.

        :param guid: New label GUID to write to the disk image
        """
        primary = self.read_primary_gpt_header()
        backup = self.read_backup_gpt_header()

        new_primary = primary.with_new_guid(guid)
        new_backup = backup.with_new_guid(guid)

        self.write_gpt_headers(new_primary, new_backup)
